// .. MACROS

#ifdef GRAPHIC_PLATFORM_Dx11
    #define TECHNIQUE technique11
    #define PS_OUTPUT_TYPE SV_Target0
    #define POSITION_INPUT_TYPE SV_POSITION
    #define VS_COMPILE_VERSION vs_4_0
    #define PS_COMPILE_VERSION ps_4_0
    #define GS_COMPILE_VERSION gs_4_0
#else
    #define TECHNIQUE technique
    #define PS_OUTPUT_TYPE COLOR0
    #define POSITION_INPUT_TYPE POSITION
    #define VS_COMPILE_VERSION vs_3_0
    #define PS_COMPILE_VERSION ps_3_0
    #define GS_COMPILE_VERSION UNSUPPORTED
#endif

// .. CONSTANTS

float4x4
    ProjectionMatrix; 
float4
    TextureModulationColor = { 1.0f, 1.0f, 1.0f, 1.0f },
    ColorParameter = { 1.0f, 1.0f, 1.0f, 1.0f };
int 
    TextureWidth,
    TextureHeight;
float3
    Offset;

// ~~

#ifdef GRAPHIC_PLATFORM_Dx11
    Texture2D
        DiffuseTexture;
#else
    Texture
        DiffuseTexture;
#endif
    
// .. TYPES

struct VS_COLORED_INPUT
{
    float3 Position : POSITION_INPUT_TYPE;
    float4 Color : COLOR0;
};

// ~~

struct VS_TEXTURED_INPUT
{
    float2 Position : POSITION_INPUT_TYPE;
    float2 TextureCoordinate : TEXCOORD0;
};

// ~~

struct VS_COLORED_OUTPUT
{
    float4 Position : POSITION_INPUT_TYPE;
    float4 Color : COLOR0;
};

// ~~

struct VS_TEXTURED_OUTPUT
{
    float4 Position : POSITION_INPUT_TYPE;
    float2 TextureCoordinate : TEXCOORD0;
};

// .. SAMPLERS

#ifdef GRAPHIC_PLATFORM_Dx11
    SamplerState DiffuseTextureSampler
    {
        Filter = MIN_MAG_MIP_LINEAR;
        AddressU = Wrap;
        AddressV = Wrap;
    };

    // ~~

    SamplerState DownsampleTextureSampler
    {
        Filter = MIN_MAG_MIP_LINEAR;
        AddressU = Wrap;
        AddressV = Wrap;
    };
    
    #define PS_SAMPLE( _TEXTURE_, _TEXTURE_COORDS_ ) \
        _TEXTURE_.Sample( _TEXTURE_##Sampler, _TEXTURE_COORDS_ )
#else
    sampler2D DiffuseTextureSampler
    { 
        Texture = <DiffuseTexture>;
        MinFilter = LINEAR;
        MagFilter = LINEAR;
        MipFilter = LINEAR;
    };

    // ~~

    sampler2D DownsampleTextureSampler
    { 
        Texture = <DiffuseTexture>; 
        MinFilter = LINEAR;
        MagFilter = LINEAR;
        MipFilter = LINEAR;
    };
    
    #define PS_SAMPLE( _TEXTURE_, _TEXTURE_COORDS_ ) \
        tex2D( _TEXTURE_##Sampler, _TEXTURE_COORDS_ )
#endif

// .. OPERATIONS

VS_TEXTURED_OUTPUT vs_textured_main(
    VS_TEXTURED_INPUT input
    )
{
    VS_TEXTURED_OUTPUT
        output;
        
    output.Position = mul( float4( input.Position, 0, 1 ), ProjectionMatrix );
    output.TextureCoordinate = input.TextureCoordinate;
    
    return output;
}

// ~~

VS_TEXTURED_OUTPUT vs_textured_main_with_offset(
    VS_TEXTURED_INPUT input
    )
{
    VS_TEXTURED_OUTPUT
        output;
        
    output.Position = mul( float4( input.Position + Offset.xy, 0, 1 ), ProjectionMatrix );
    output.TextureCoordinate = input.TextureCoordinate;
    
    return output;
}

// ~~

VS_TEXTURED_OUTPUT vs_textured_no_projection_main(
    VS_TEXTURED_INPUT input
    )
{
    VS_TEXTURED_OUTPUT
        output;
        
    output.Position = float4( input.Position, 0, 1 );
    output.TextureCoordinate = input.TextureCoordinate;
    
    return output;
}

// ~~

VS_COLORED_OUTPUT vs_colored_main(
    VS_COLORED_INPUT input
    )
{
    VS_COLORED_OUTPUT
        output;

    output.Position = mul( float4( input.Position, 1 ), ProjectionMatrix );
    output.Color = input.Color;
    
    return output;
}

// ~~

float4 ps_colored_main(
    #ifdef GRAPHIC_PLATFORM_Dx11
        float4 Position : POSITION_INPUT_TYPE,
    #endif
    float4 color : COLOR0
    ) : PS_OUTPUT_TYPE
{
    return TextureModulationColor * color;
}

// ~~

float4 ps_textured_main(
    #ifdef GRAPHIC_PLATFORM_Dx11
        float4 Position : POSITION_INPUT_TYPE,
    #endif
    float2 texture_coordinate : TEXCOORD0 
    ) : PS_OUTPUT_TYPE
{
    return PS_SAMPLE( DiffuseTexture, texture_coordinate ); 
}

// ~~

float4 ps_textured_main_with_blend(
    #ifdef GRAPHIC_PLATFORM_Dx11
        float4 Position : POSITION_INPUT_TYPE,
    #endif
    float2 texture_coordinate : TEXCOORD0 
    ) : PS_OUTPUT_TYPE
{
    return ColorParameter * PS_SAMPLE( DiffuseTexture, texture_coordinate ); 
}

// ~~

float4 ps_textured_main_with_blend_to_color(
    #ifdef GRAPHIC_PLATFORM_Dx11
        float4 Position : POSITION_INPUT_TYPE,
    #endif
   float2 texture_coordinate : TEXCOORD0 
   ) : PS_OUTPUT_TYPE
{
    float4
        color;
    
    color = PS_SAMPLE( DiffuseTexture, texture_coordinate );
    color.xyz = lerp( color.xyz, ColorParameter.xyz, 1.0f - ColorParameter.w );
    
    return color;
}

// ~~

float4 ps_downsample_main(
    #ifdef GRAPHIC_PLATFORM_Dx11
        float4 Position : POSITION_INPUT_TYPE,
    #endif
    float2 texture_coordinate : TEXCOORD0 
    ) : PS_OUTPUT_TYPE
{
    float 
        offset_x,
        offset_y;
    float4
        color[4],
        result;
        
    offset_x = 1.0 / TextureWidth;
    offset_y = 1.0 / TextureHeight;
    
    texture_coordinate.x += offset_x * 0.5;
    texture_coordinate.y += offset_y * 0.5 ;
    
    color[ 0 ] = PS_SAMPLE( DiffuseTexture, texture_coordinate + float2( offset_x, offset_y ) ); 
    color[ 1 ] = PS_SAMPLE( DiffuseTexture, texture_coordinate + float2( offset_x, 0 ) ); 
    color[ 2 ] = PS_SAMPLE( DiffuseTexture, texture_coordinate + float2( 0, 0 ) ); 
    color[ 3 ] = PS_SAMPLE( DiffuseTexture, texture_coordinate + float2( 0, offset_y ) );
    
    result.rgb = color[0].rgb * color[0].a + color[1].rgb * color[1].a + color[2].rgb * color[2].a + color[3].rgb * color[3].a;
    result.a = color[0].a + color[1].a + color[2].a + color[3].a;
    result.rgb /= result.a;
    result.a /= 4.0;
    
    return result;
}

// .. TECHNIQUES

TECHNIQUE ColoredInput
{
    pass P0
    {
        VertexShader = compile VS_COMPILE_VERSION vs_colored_main();
        PixelShader = compile PS_COMPILE_VERSION ps_colored_main();
    }
}

// ~~

TECHNIQUE TexturedInputWithBlendToColor
{
    pass P0
    {
        VertexShader = compile VS_COMPILE_VERSION vs_textured_main();
        PixelShader = compile PS_COMPILE_VERSION ps_textured_main_with_blend_to_color();
    }
}

// ~~

TECHNIQUE TexturedInputWithBlend
{
    pass P0
    {
        VertexShader = compile VS_COMPILE_VERSION vs_textured_main();
        PixelShader = compile PS_COMPILE_VERSION ps_textured_main_with_blend();
    }
}

// ~~

TECHNIQUE TexturedInput
{
    pass P0
    {
        VertexShader = compile VS_COMPILE_VERSION vs_textured_main();
        PixelShader = compile PS_COMPILE_VERSION ps_textured_main();
    }
}

// ~~

TECHNIQUE TexturedInputWithOffset
{
    pass P0
    {
        VertexShader = compile VS_COMPILE_VERSION vs_textured_main_with_offset();
        PixelShader = compile PS_COMPILE_VERSION ps_textured_main();
    }
}

// ~~

TECHNIQUE DownSample
{
    pass P0
    {
        VertexShader = compile VS_COMPILE_VERSION vs_textured_main();
        PixelShader = compile PS_COMPILE_VERSION ps_downsample_main();
    }
}

// ~~

TECHNIQUE TexturedInputNoProjection
{
    pass P0
    {
        VertexShader = compile VS_COMPILE_VERSION vs_textured_no_projection_main();
        PixelShader = compile PS_COMPILE_VERSION ps_textured_main();
    }
}