// .. CONSTANTS

shared float4x4
    InverseProjectionMatrix,
    WorldToViewXf,
    WvpXf;
shared float2
    ScreenSize;

texture2D NormalTexture;
sampler2D NormalTextureSampler = sampler_state
{
    Texture = <NormalTexture>;

    MinFilter = LINEAR;
    MagFilter = LINEAR;
    MipFilter = LINEAR;
};

texture2D DepthTexture;
sampler2D DepthTextureSampler = sampler_state
{
    Texture = <DepthTexture>;

    MinFilter = LINEAR;
    MagFilter = LINEAR;
    MipFilter = LINEAR;
};

// .. TYPES

struct VS_POINT_LIGHT_INPUT
{
    float3 Position : POSITION;
    float3 LightPosition : TEXCOORD0;
    float3 LightParameter : TEXCOORD1;
    float3 Color : COLOR0;
};

struct VS_POINT_LIGHT_OUTPUT
{
    float4 Position         : POSITION;
    float3 ViewSpaceLightPosition    : TEXCOORD0;
    float3 LightParameter   : TEXCOORD1;
    float3 Color            : COLOR0;
};

struct PS_INPUT
{
    float3 ViewSpaceLightPosition    : TEXCOORD0;
    float3 LightParameter   : TEXCOORD1;
    float3 Color            : COLOR0;
};

// ~~

VS_POINT_LIGHT_OUTPUT vs_main(
    VS_POINT_LIGHT_INPUT input
    )
{
    VS_POINT_LIGHT_OUTPUT
        output;

    output.Position = mul( float4( input.Position, 1 ), WvpXf );
    output.ViewSpaceLightPosition = mul( float4( input.LightPosition, 1 ), WorldToViewXf ).xyz;
    output.LightParameter = input.LightParameter;
    output.Color = input.Color;
    
    return output;
}

// ~~

float luminosity( float3 color )
{
    return dot( color, float3( 0.2126, 0.7152, 0.0722 ) );    
}

// ~~

float LOCAL_GetBlinnSpecular(
    const float3 view_direction,
    const float3 light_direction,
    const float3 normal
    )
{
    float3
        half_vector;
    float
        specular;
    
    half_vector = normalize( view_direction + light_direction );
    specular = saturate( dot( half_vector, normal ) );
    
    return specular;
}

// ~~
    
float4 ps_main(
    PS_INPUT input,
    float2 vpos : VPOS
    ) : COLOR0
{
    float
        view_space_depth,
        distance,
        a,
        b,
        attenuation,
        n_dot_l,
        specular_intensity;
    float2
        screen_uv,
        corrected_screen_uv;
    float3
        normal,
        light_direction,
        view;
    float4
        view_space_position,
        result;
    
    screen_uv = vpos;
    screen_uv /= ScreenSize;
    corrected_screen_uv = screen_uv + 0.5f / ScreenSize;

    normal.xy = -1.0f + 2.0f * tex2D( NormalTextureSampler, corrected_screen_uv ).rg;
    normal.z = sqrt( 1.0f - normal.x * normal.x - normal.y * normal.y );

    view_space_depth = tex2D( DepthTextureSampler, corrected_screen_uv ).r;

    screen_uv *= 2.0f;
    screen_uv -= 1.0f;
    view_space_position = mul( float4( screen_uv.x, -screen_uv.y, view_space_depth, 1.0f ), InverseProjectionMatrix );
    view_space_position.xyz /= view_space_position.w;

    light_direction = input.ViewSpaceLightPosition.xyz - view_space_position.xyz;
    
    distance = length( light_direction );
    a = 1.0 / ( input.LightParameter.x - input.LightParameter.y );
    b = - a * input.LightParameter.y;
    
    //:TODO: use decay type ( input.LightParameter.z )
    attenuation = saturate( a * distance + b );
    
    light_direction /= distance;
    n_dot_l = dot( normal, light_direction );
    result.rgb = attenuation * saturate( n_dot_l ) * input.Color;

    result.a = step( 0.0, n_dot_l );
    result.a *= luminosity( input.Color );
    result.a *= pow( abs( attenuation ), 0.25f ); //:TODO: parametrize this specular attenuation exponent
    result.a *= LOCAL_GetBlinnSpecular( normalize( -view_space_position.xyz ), light_direction, normal ) ;

    result.a = saturate( pow( abs( result.a ), 10.0f ) );

    return result;
}

// ~~

technique Default
{
    pass P0
    {
        VertexShader = compile vs_3_0 vs_main();
        PixelShader = compile ps_3_0 ps_main();
    }
}
