// .. MACROS

#ifdef GRAPHIC_PLATFORM_Dx11
    #define TECHNIQUE technique11
    #define PS_OUTPUT_TYPE SV_Target0
    #define POSITION_INPUT_TYPE SV_POSITION
    #define VS_COMPILE_VERSION vs_4_0
    #define PS_COMPILE_VERSION ps_4_0
    #define GS_COMPILE_VERSION gs_4_0
#else
    #define TECHNIQUE technique
    #define PS_OUTPUT_TYPE COLOR0
    #define POSITION_INPUT_TYPE POSITION
    #define VS_COMPILE_VERSION vs_3_0
    #define PS_COMPILE_VERSION ps_3_0
    #define GS_COMPILE_VERSION UNSUPPORTED
#endif

// .. CONSTANTS

float2
    Position = { 0.0, 0.0 },
    ScreenSize;
float4
    TextureModulationColor = { 1.0, 1.0, 1.0, 1.0 };
float4x4
    ProjectionMatrix,
    ModelViewMatrix;

// ~~

#ifdef GRAPHIC_PLATFORM_Dx11
    Texture2D
        DiffuseTexture,
        MaskTexture;
#else
    Texture
        DiffuseTexture,
        MaskTexture;
#endif

// .. TYPES

struct VS_TEXTURED_INPUT
{
    float2 Position : POSITION_INPUT_TYPE;
    float2 TextureCoordinate : TEXCOORD0;
};

// ~~

struct VS_POSITION_INPUT
{
    float2 Position : POSITION_INPUT_TYPE;
};

// ~~

struct VS_TEXTURED_OUTPUT
{
    float4 Position : POSITION_INPUT_TYPE;
    float2 TextureCoordinate : TEXCOORD0;
};

// ~~

struct VS_POSITION_OUTPUT
{
    float4 Position : POSITION_INPUT_TYPE;
};

// .. SAMPLERS

#ifdef GRAPHIC_PLATFORM_Dx11
    SamplerState DiffuseTextureSampler
    {
        Filter = MIN_MAG_MIP_LINEAR;
        AddressU = Wrap;
        AddressV = Wrap;
    };

    // ~~

    SamplerState MaskTextureSampler
    {
        Filter = MIN_MAG_MIP_LINEAR;
        AddressU = Wrap;
        AddressV = Wrap;
    };
    
    #define PS_SAMPLE( _TEXTURE_, _TEXTURE_COORDS_ ) \
        _TEXTURE_.Sample( _TEXTURE_##Sampler, _TEXTURE_COORDS_ )
#else
    sampler2D DiffuseTextureSampler
    { 
        Texture = <DiffuseTexture>;
        MinFilter = LINEAR;
        MagFilter = LINEAR;
        MipFilter = LINEAR;
    };

    // ~~

    sampler2D MaskTextureSampler
    { 
        Texture = <MaskTexture>;
        MinFilter = LINEAR;
        MagFilter = LINEAR;
        MipFilter = LINEAR;
    };

    #define PS_SAMPLE( _TEXTURE_, _TEXTURE_COORDS_ ) \
        tex2D( _TEXTURE_##Sampler, _TEXTURE_COORDS_ )
#endif
    
// .. OPERATIONS

VS_TEXTURED_OUTPUT vs_main(
    VS_TEXTURED_INPUT input
    )
{
    VS_TEXTURED_OUTPUT
        output;
        
    output.Position = float4( 2.0 * ( input.Position + Position ) / ScreenSize - 1.0, 0, 1 );
    output.TextureCoordinate = input.TextureCoordinate;
    
    return output;
}

// ~~

float4 ps_main(
    #ifdef GRAPHIC_PLATFORM_Dx11
        float4 pos : POSITION_INPUT_TYPE,
    #endif
    float2 texture_coordinate : TEXCOORD0,
    uniform const bool use_diffuse_texture,
    uniform const bool use_mask
   ) : PS_OUTPUT_TYPE
{
    float4
        result;
    
    result = TextureModulationColor;

    if ( use_diffuse_texture )
    {
        result *= PS_SAMPLE( DiffuseTexture, texture_coordinate );
    }

    if ( use_mask )
    {
        result.a *= PS_SAMPLE( MaskTexture, texture_coordinate ).r;
    }

    return result;
}

// ~~

float4 ps_texture_none_main(
    float4 pos : POSITION_INPUT_TYPE
    ) : PS_OUTPUT_TYPE
{
    return TextureModulationColor;
}
// ~~

VS_POSITION_OUTPUT vs_texture_none_main(
    VS_POSITION_INPUT input
    )
{
    VS_POSITION_OUTPUT
        output;
        
    output.Position = mul( float4( input.Position, 0, 1 ), ProjectionMatrix );

    return output;
}

// ~~

VS_TEXTURED_OUTPUT vs_matrix_main(
    VS_TEXTURED_INPUT input
    )
{
    VS_TEXTURED_OUTPUT
        output;
    float4x4
        wvp;

    wvp = mul( ModelViewMatrix, ProjectionMatrix );
        
    output.Position = mul( float4( input.Position, 0, 1 ), wvp ) ;
    output.TextureCoordinate = input.TextureCoordinate;    
    return output;
}

// .. TECHNIQUES

TECHNIQUE Default
{
    pass P0
    {
        VertexShader = compile VS_COMPILE_VERSION vs_main();
        PixelShader = compile PS_COMPILE_VERSION ps_main( true, false );
    }
}

// ~~

TECHNIQUE UseMatrix
{
    pass P0
    {
        VertexShader = compile VS_COMPILE_VERSION vs_matrix_main();
        PixelShader = compile PS_COMPILE_VERSION ps_main( true, false );
    }
}

// ~~

TECHNIQUE Mask
{
    pass P0
    {
        VertexShader = compile VS_COMPILE_VERSION vs_matrix_main();
        PixelShader = compile PS_COMPILE_VERSION ps_main( true, true );
    }
}

// ~~

TECHNIQUE UseMatrixNoTexture
{
    pass P0
    {
        VertexShader = compile VS_COMPILE_VERSION vs_matrix_main();
        PixelShader = compile PS_COMPILE_VERSION ps_main( false, false );
    }
}

// ~~

TECHNIQUE TechniqueNoTexture
{
    pass P0
    {
        VertexShader = compile VS_COMPILE_VERSION vs_main();
        PixelShader = compile PS_COMPILE_VERSION ps_main( true, false );
    }
}

// ~~

TECHNIQUE TextureNone
{
    pass P0
    {
        VertexShader = compile VS_COMPILE_VERSION vs_texture_none_main();
        PixelShader = compile PS_COMPILE_VERSION ps_texture_none_main();
    }
}
